from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordUISerializerBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_ui_serializer"
    section = "json-serializer"
    template = "ui-serializer"



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import (
    convert_config_to_qualified_name,
    package_name,
)
from oarepo_model_builder.validation.utils import ImportSchema

from .service import ServiceModelComponent
from .utils import set_default


class RecordItemClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(metadata={"doc": "Generate the record item class"})
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the record item class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "List of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be put below the record item class"},
    )
    components = ma.fields.List(
        ma.fields.String(), metadata={"doc": "List of record item components"}
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()


class RecordItemModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [ServiceModelComponent]

    class ModelSchema(ma.Schema):
        record_item = ma.fields.Nested(
            RecordItemClassSchema,
            attribute="record-item",
            data_key="record-item",
            metadata={"doc": "Record item class settings"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        record_prefix = datatype.definition["module"]["prefix"]

        service_package = package_name(datatype.definition["service-config"]["module"])

        record_item_config = set_default(datatype, "record-item", {})

        record_item_config.setdefault("generate", True)
        record_item_module = record_item_config.setdefault(
            "module",
            f"{service_package}.results",
        )
        record_item_config.setdefault(
            "class",
            f"{record_item_module}.{record_prefix}RecordItem",
        )
        record_item_config.setdefault("extra-code", "")
        record_item_config.setdefault(
            "base-classes",
            [
                "oarepo_runtime.services.results.RecordItem",
            ],
        )
        record_item_config.setdefault("components", [])
        record_item_config.setdefault(
            "imports",
            [],
        )
        convert_config_to_qualified_name(record_item_config)



from collections import defaultdict
from typing import List

from ..datatypes.components.facets import FacetDefinition
from .invenio_base import InvenioBaseClassPythonBuilder
from .invenio_record_facets import get_distinct_facets


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search_options"
    section = "search-options"
    template = "record-search-options"

    def finish(self, **extra_kwargs):
        facets: List[FacetDefinition] = get_distinct_facets(self.current_model)
        facet_groups = {}
        default_group = []
        search_data = []

        # gather all facet group names
        facet_group_names = set()
        for f in facets:
            facet_group_names.update(f.facet_groups.keys())

        # there might be a definition of facet groups on the model's facets -> facet-groups
        # if there is, we need to merge those
        top_level_facet_groups = self.current_model.definition.get("facets", {}).get(
            "facet-groups", {}
        )
        facet_group_names.update(top_level_facet_groups.keys())
        path_to_facet_group = defaultdict(dict)

        for group, group_def in top_level_facet_groups.items():
            for path, priority in group_def.items():
                path_to_facet_group[group][path] = priority

        # for each group name, gather all facets with that group name
        # sort them by their order in the group
        # and add them to the facet_groups dict
        for group in sorted(facet_group_names):
            # skip the default group
            if group == "_default":
                continue

            # gather all facets with this facet_group_name
            group_members = []
            for f in facets:
                if group in f.facet_groups:
                    group_members.append((f, f.facet_groups[group]))

                # if there is a definition of facet groups on the model's facets -> add the selected facets as well
                for path, priority in path_to_facet_group.get(group, {}).items():
                    if f.dot_path.startswith(path):
                        group_members.append((f, priority))

            # sort the group members by their order in the facet_group
            group_members.sort(key=lambda x: x[1])

            if group not in facet_groups.keys():
                facet_groups[group] = {}

            for f, _ in group_members:
                facet_groups[group][f.path] = "facets." + f.path

        for f in facets:
            if len(f.facet_groups) > 0:
                default_group.append({f.path: "facets." + f.path})
            search_data.append({f.path: "facets." + f.path})

        if "sortable" in self.current_model.definition:
            sort_options = self.current_model.definition["sortable"]
        else:
            sort_options = {}

        extra_kwargs["facet_groups"] = facet_groups
        extra_kwargs["default_group"] = default_group
        extra_kwargs["sort_definition"] = sort_options
        super().finish(**extra_kwargs)



import copy
from typing import Union


def deepmerge(
    target,
    source,
    stack=None,
    listmerge: Union[str, callable] = "overwrite",
    dictmerge=None,
):
    if stack is None:
        stack = []

    if isinstance(target, dict):
        if source is not None:
            if not isinstance(source, dict):
                raise AttributeError(
                    f"Incompatible source and target on path {stack}: source {source}, target {target}"
                )
            if dictmerge:
                merged = dictmerge(target, source, stack)
            else:
                merged = None
            if merged is None:
                for k, v in source.items():
                    if k not in target:
                        target[k] = source[k]
                    else:
                        target[k] = deepmerge(
                            target[k],
                            source[k],
                            stack + [k],
                            listmerge=listmerge,
                            dictmerge=dictmerge,
                        )
    elif isinstance(target, list):
        if source is not None:
            if not isinstance(source, list):
                raise AttributeError(
                    f"Incompatible source and target on path {stack}: source {source}, target {target}"
                )
            if listmerge == "overwrite":
                for idx in range(min(len(source), len(target))):
                    target[idx] = deepmerge(
                        target[idx],
                        source[idx],
                        stack + [idx],
                        listmerge=listmerge,
                        dictmerge=dictmerge,
                    )
                for idx in range(len(target), len(source)):
                    target.append(source[idx])
            elif listmerge == "extend":
                target.extend(source)
            elif listmerge == "keep":
                if len(source) > len(target):
                    target.extend(source[len(target) :])
            elif callable(listmerge):
                listmerge(target, source, stack)
            else:
                raise AttributeError(
                    'listmerge must be one of "overwrite", "extend" or "keep"'
                )
    elif target is None:
        return copy.deepcopy(source)
    return target



from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioRecordMetadataModelsSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_metadata_models_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        output.add_entry_point(
            "invenio_db.models",
            self.current_model.definition["record-metadata"]["alias"],
            self.current_model.definition["record-metadata"]["module"],
        )



from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioRecordJSONSchemasSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_jsonschemas_setup_cfg"

    def finish(self):
        super().finish()
        if self.current_model.definition["json-schema-settings"].get("skip"):
            return

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        output.add_entry_point(
            "invenio_jsonschemas.schemas",
            self.current_model.definition["json-schema-settings"]["alias"],
            self.current_model.definition["json-schema-settings"]["module"],
        )



import os
from pathlib import Path

from ..datatypes import Section
from ..outputs.cfg import CFGOutput
from ..utils.dict import dict_get
from . import OutputBuilder
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules


class ModelSaverBuilder(JSONBaseBuilder):
    TYPE = "model_saver"
    output_file_type = "json"
    output_file_name = ["saved-model", "file"]
    parent_module_root_name = "models"

    def build_node(self, node):
        generated = self.generate(node)
        self.output.merge({"model": generated})

    def generate(self, node):
        section: Section = node.section_model_saver
        ret = {**node.definition}
        ret.pop("properties", None)
        ret.pop("items", None)
        ret.update(**section.config)

        if section.children:
            properties = ret.setdefault("properties", {})
            for k, v in section.children.items():
                v = self.generate(v)
                properties[k] = v
        if section.item:
            ret["items"] = self.generate(section.item)
        return ret

    def begin(self, current_model, schema):
        super().begin(current_model, schema)

        output_name = dict_get(self.current_model.definition, self.output_file_name)
        self.output = self.builder.get_output(self.output_file_type, output_name)
        ensure_parent_modules(
            self.builder, Path(output_name), ends_at=self.parent_module_root_name
        )


class ModelRegistrationBuilder(OutputBuilder):
    TYPE = "model_registration"

    def finish(self):
        super().finish()
        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")
        module = self.current_model.definition["saved-model"]["module"]
        output.add_entry_point(
            "oarepo.models",
            self.current_model.definition["saved-model"]["alias"],
            f"{module}:{os.path.basename(self.current_model.definition['saved-model']['file'])}",
        )



from collections import defaultdict
from typing import Dict

import marshmallow as ma

from oarepo_model_builder.datatypes import ModelDataType, datatypes
from oarepo_model_builder.utils.dict import dict_get
from oarepo_model_builder.utils.python_name import (
    convert_config_to_qualified_name,
    parent_module,
)

from ..marshmallow import ObjectMarshmallowComponent, ObjectMarshmallowExtraSchema
from .defaults import DefaultsModelComponent
from .service import ServiceModelComponent
from .utils import set_default


class ModelMarshmallowSchema(ObjectMarshmallowExtraSchema):
    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
    )


class MarshmallowModelMixin:
    model_marshmallow_section = None
    context_registered_classes_name = None
    register_class_names_method = None
    build_class_names_method = None

    def after_model_prepare(self, *, datatype, **__kwargs):
        classes = defaultdict(list)
        marshmallow_def = dict_get(datatype.definition, self.model_marshmallow_section)
        marshmallow_module = marshmallow_def["module"]

        for node in datatype.deep_iter():
            datatypes.call_components(
                node,
                self.register_class_names_method,
                classes=classes,
                marshmallow_module=marshmallow_module,
            )
        for node in datatype.deep_iter():
            datatypes.call_components(
                node,
                self.build_class_names_existing_method,
                classes=classes,
                marshmallow_module=marshmallow_module,
            )
        for node in datatype.deep_iter():
            datatypes.call_components(
                node,
                self.build_class_names_new_method,
                classes=classes,
                marshmallow_module=marshmallow_module,
            )


class MarshmallowModelComponent(MarshmallowModelMixin, ObjectMarshmallowComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, ServiceModelComponent]

    model_marshmallow_section = ["marshmallow"]
    context_registered_classes_name = "marshmallow-classes"
    register_class_names_method = "marshmallow_register_class_name"
    build_class_names_existing_method = "marshmallow_build_class_name_existing"
    build_class_names_new_method = "marshmallow_build_class_name_new"

    class ModelSchema(ma.Schema):
        marshmallow = ma.fields.Nested(ModelMarshmallowSchema)

    def marshmallow_register_class_names(self, *, datatype, classes, **kwargs):
        marshmallow_def = dict_get(datatype.definition, self.model_marshmallow_section)
        classes[marshmallow_def["class"]].append((True, datatype))

    def before_model_prepare(self, datatype, *, context, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
        services_module = parent_module(datatype.definition["service"]["module"])

        marshmallow: Dict = set_default(datatype, "marshmallow", {})
        marshmallow.setdefault("generate", True)
        module = marshmallow.setdefault("module", f"{services_module}.schema")
        marshmallow.setdefault("class", f"{module}.{prefix}Schema")
        marshmallow.setdefault("extra-code", "")
        marshmallow.setdefault("base-classes", ["marshmallow.Schema"])
        convert_config_to_qualified_name(marshmallow)

        if "properties" in datatype.definition and "metadata" in (
            datatype.definition["properties"] or {}
        ):
            metadata_marshmallow = set_default(
                datatype, "properties", "metadata", "marshmallow", {}
            )
            metadata_module = metadata_marshmallow.setdefault("module", module)
            metadata_marshmallow.setdefault("generate", True)
            metadata_marshmallow.setdefault(
                "class", f"{metadata_module}.{prefix}MetadataSchema"
            )
            metadata_marshmallow.setdefault("extra-code", "")
            metadata_marshmallow.setdefault("base-classes", ["marshmallow.Schema"])
            convert_config_to_qualified_name(metadata_marshmallow)



from .object import ObjectDataType


class NestedDataType(ObjectDataType):
    model_type = "nested"

    mapping = {"type": "nested"}
    facets = {
        "facet-class": "oarepo_runtime.services.facets.nested_facet.NestedLabeledFacet",
    }



from ..datatypes import DataType


class FlatObjectDataType(DataType):
    model_type = "flat_object"

    ui = {
        "marshmallow": {
            "field-class": "marshmallow.fields{ma_fields.Dict}",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.Dict}",
    }
    json_schema = {"type": "object"}
    mapping = {"type": "flat_object"}



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordItemBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_item"
    section = "record-item"
    template = "record-item"



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType

from .defaults import DefaultsModelComponent
from .record import RecordModelComponent


class SortSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    definition = ma.fields.String(required=False)


class SortModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, RecordModelComponent]

    class ModelSchema(ma.Schema):
        sortable = ma.fields.Nested(SortSchema, required=False)

    def before_model_prepare(self, datatype, *, context, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
